import { useUserStore } from '/@/store/modules/user'
import { defHttp, otherHttp } from '/@/utils/http/axios'

enum Api {
    GetVMId = 'mgr/get_vids',
    GetVMPid = 'port/get_pids',
    GetVMRAM = 'port/get_mem',
    GetVMStack = 'port/get_pid_stack',
    GetVMLoad = 'mgr/get_load'
}

interface Pid {
    vid: number,
}

interface PidStack {
    vid: number,
    cid: number
}

export function getUserInfoParams() {
    const userStore = useUserStore()
    let userInfo = userStore.getUserInfo
    let userInfoParams = {
        user: userInfo.user as any,
        token: userInfo.token as any
    }
    return userInfoParams;
}

/**
 * @description: 获取vm
 */
export function getVMId() {
    return otherHttp.get(
        { url: Api.GetVMId, headers: { 'Authorization': 'Basic ' + btoa(getUserInfoParams().user + ':' + getUserInfoParams().token) } }, { errorMessageMode: 'none' }
    )
}

/**
 * @description: 获取vm pid
 */
export function getVMPid(params: Pid) {
    return otherHttp.get(
        { url: Api.GetVMPid, params, headers: { 'Authorization': 'Basic ' + btoa(getUserInfoParams().user + ':' + getUserInfoParams().token) } }, { errorMessageMode: 'none' }
    )
}


/**
 * @description: 获取vm 内存
 */
export function getVMRAM(params: Pid) {
    return otherHttp.get(
        { url: Api.GetVMRAM, params, headers: { 'Authorization': 'Basic ' + btoa(getUserInfoParams().user + ':' + getUserInfoParams().token) } }, { errorMessageMode: 'none' }
    )
}

/**
 * @description: 获取vm pid 堆栈
 */
export function getVMStack(params: PidStack) {
    return otherHttp.get(
        { url: Api.GetVMStack, params, headers: { 'Authorization': 'Basic ' + btoa(getUserInfoParams().user + ':' + getUserInfoParams().token) } }, { errorMessageMode: 'none' }
    )
}

/**
 * @description: 获取vm pid负载
 */
export function getVMLoad() {
    return otherHttp.get(
        { url: Api.GetVMLoad, headers: { 'Authorization': 'Basic ' + btoa(getUserInfoParams().user + ':' + getUserInfoParams().token) } }, { errorMessageMode: 'none' }
    )
}
