/**
 * **注：** 这个文件被构建系统特殊处理，会放到sys.js的最开头。
 * 
 * 模块加载，主要是 3个 接口：
 * 
 *   _$pi.define("a", ["require", "exports", "./b", "./c"], (require, exports, b, c) => { })
 *   
 *   _$pi.require(["./a", "./b"], (a, b) => {...})
 * 
 *   try {
 *          let a = _$pi.import("./a");
 *      或：
 *          let [a, b] = _$pi.import(["./a", "./b"]);
 *   } catch (e) {
 *      console.log("取不到模块");
 *   }
 */

/**
 * ## 关于 Wasm 的约定
 * 
 * 为了兼容所有平台，在同一个目录下：
 * 
 * - wasm本身的代码，命名为：***.wasm，比如：gui.wasm
 * - wasm封装的js/ts代码，前面和wasm同名，比如：gui.wasm.ts/js
 * - 每个wasm必须有个同名的asm.js，比如：gui.asm.js
 * 
 * ## wasm 代码加载-模块加载 策略：
 * 
 *     - 模块引用只能引用 .wasm.js 模块
 *     - wasm 功能判断，全局变量有 相关字段；
 *     - 加载：同目录下同名的 wasm.js；如果不支持wasm，会抛异常
 *     - wasm.js 代码处理
 *          + 加载同名的wasm资源文件，加载-编译成功，拿到WsamModule
 *          + 加载同名的wasm.js模块
 *          + 模块系统，需要等待：wasm.js 模块 的 postRun() 调用；
 *          + wasm.js模块 加载完成；
 */

// 只能用var
console.log("pisys start");
var self;
try {
	self = window;
	window.self = self;
} catch (e) {

}

if (!self) {
	try {
		self = global;
		global.self = self;
		global.window = self;
	} catch (e) {

	}
}

if (!self) {
	try {
		self = GameGlobal;
		GameGlobal.self = self;
		GameGlobal.global = self;
		global.window = self;
	} catch (e) {

	}
}

try {

	if (!self.GameGlobal) {
		self.GameGlobal = self;
	}

	if (!self.global) {
		self.global = self;
	}

	if (!self.window) {
		self.window = self;
	}

	// chrome 71 及其以上才支持 globalThis
	if (!self.globalThis) {
		self.globalThis = self;
	}
} catch (error) {

}

if (!self) {
	try {
		self = globalThis;
		globalThis.self = self;
		globalThis.global = self;
		global.window = self;
	} catch (e) {

	}
}

// 小游戏平台
if (!self.navigator || self.navigator.userAgent.indexOf("wechatdevtools") >= 0) {
	// console.log("小游戏平台：", self.navigator);
	console.log("minigame platform：", self.navigator);
	let userAgent = "PI_WX_GAME";
	try {
		let _ = !!qq;
		// QQ小游戏
		userAgent = "PI_QQ_GAME";
	} catch (_) {
		// TODO：测试别的小游戏平台；
	}

	if (!self.navigator) {
		self.navigator = {
			userAgent,
		}
	}
}

// 支持Promise.any  chrome 85才原生支持 
Promise && !Promise.any && (Promise.any = o => new Promise((i, n) => { var e, l, t, u, d; let r = !1, c = [], v = 0, s = []; function f(o) { r || (r = !0, i(o)) } function h(o) { s.push(o), s.length >= v && n(s) } for (let i of o) v++, c.push(i); for (let o of c) void 0 !== (null === (e = o) || void 0 === e ? void 0 : e.then) || void 0 !== (null === (l = o) || void 0 === l ? void 0 : l.catch) ? (null === (u = null === (t = o) || void 0 === t ? void 0 : t.then(o => f(o))) || void 0 === u || u.catch(o => { }), null === (d = o) || void 0 === d || d.catch(o => h(o))) : f(o) }));

(function () {
	/**
	 * @description utf8的Uint8Array解码成字符串
	 * @example
	 */
	const utf8Decode = (_arr) => {
		if (!_arr || _arr.byteLength === 0) {
			return '';
		}
		var arr = null;
		if (_arr instanceof ArrayBuffer) {
			arr = new Uint8Array(_arr);
		} else {
			arr = _arr;
		}
		var c;
		var out = '';
		var i = 0;
		var len = arr.length;
		while (i < len) {
			c = arr[i++];
			if (c < 128) {
				out += String.fromCharCode(c);
			} else if (c < 0xE0 && i < len) {
				out += String.fromCharCode(((c & 0x1F) << 6) | (arr[i++] & 0x3F));
			} else if (c < 0xF0 && i + 1 < len) {
				out += String.fromCharCode((((c & 0x0F) << 12) | ((arr[i++] & 0x3F) << 6) | (arr[i++] & 0x3F)));
			} else if (c < 0xF8 && i + 2 < len) {
				out += String.fromCharCode((((c & 0x07) << 18) | ((arr[i++] & 0x3F) << 12) | ((arr[i++] & 0x3F) << 6) | (arr[i++] & 0x3F)));
			} else if (c < 0xFC && i + 3 < len) {
				// tslint:disable:max-line-length
				out += String.fromCharCode((((c & 0x03) << 24) | ((arr[i++] & 0x3F) << 18) | ((arr[i++] & 0x3F) << 12) | ((arr[i++] & 0x3F) << 6) | (arr[i++] & 0x3F)));
			} else if (c < 0xFE && i + 4 < len) {
				out += String.fromCharCode((((c & 0x01) << 30) | ((arr[i++] & 0x3F) << 24) | ((arr[i++] & 0x3F) << 18) | ((arr[i++] & 0x3F) << 12) | ((arr[i++] & 0x3F) << 6) | (arr[i++] & 0x3F)));
			} else {
				throw new Error('invalid utf8');
			}
		}

		return out;
	};

	/**
	 * @description 字符串编码成utf8的Uint8Array
	 * @example
	 */
	const utf8Encode = (s) => {
		if ((!s) || s.length === 0) {
			return new Uint8Array([]);
		}

		var bytes = [];
		var leadSurrogate = null;
		for (var i = 0; i < s.length; i++) {
			var codePoint = s.charCodeAt(i);
			// is surrogate component
			if (codePoint > 0xD7FF && codePoint < 0xE000) {
				// last char was a lead
				if (leadSurrogate) {
					// 2 leads in a row
					if (codePoint < 0xDC00) {
						bytes.push(0xEF, 0xBF, 0xBD);
						leadSurrogate = codePoint;
						continue
					} else {
						// valid surrogate pair
						codePoint = leadSurrogate - 0xD800 << 10 | codePoint - 0xDC00 | 0x10000;
						leadSurrogate = null;
					}
				} else {
					// no lead yet

					if (codePoint > 0xDBFF) {
						// unexpected trail
						bytes.push(0xEF, 0xBF, 0xBD);
						continue;
					} else if (i + 1 === s.length) {
						// unpaired lead
						bytes.push(0xEF, 0xBF, 0xBD);
						continue;
					} else {
						// valid lead
						leadSurrogate = codePoint;
						continue;
					}
				}
			} else if (leadSurrogate) {
				// valid bmp char, but last char was a lead
				bytes.push(0xEF, 0xBF, 0xBD);
				leadSurrogate = null;
			}

			// encode utf8
			if (codePoint < 0x80) {
				bytes.push(codePoint);
			} else if (codePoint < 0x800) {
				bytes.push(
					codePoint >> 0x6 | 0xC0,
					codePoint & 0x3F | 0x80
				);
			} else if (codePoint < 0x10000) {
				bytes.push(
					codePoint >> 0xC | 0xE0,
					codePoint >> 0x6 & 0x3F | 0x80,
					codePoint & 0x3F | 0x80
				);
			} else if (codePoint < 0x200000) {
				bytes.push(
					codePoint >> 0x12 | 0xF0,
					codePoint >> 0xC & 0x3F | 0x80,
					codePoint >> 0x6 & 0x3F | 0x80,
					codePoint & 0x3F | 0x80
				);
			} else {
				throw new Error('Invalid code point');
			}
		}
		return new Uint8Array(bytes);
	};

	if (!window.TextDecoder) {
		window.TextDecoder = function (flag, data) {
			if (globalThis.utf8ToStr) {
				this.decode = function (data) {
					if (!data || data.byteLength === 0) {
						return '';
					}
					if (ArrayBuffer.isView(data)) {
						if (data.byteOffset === 0 && data.byteLength === data.buffer.byteLength) {
							data = data.buffer;
						} else {
							data = data.buffer.slice(data.byteOffset, data.byteOffset + data.byteLength);
						}
					}

					return globalThis.utf8ToStr(data);
				}
			} else {
				this.decode = function (data) {
					return utf8Decode(data);
				}
			}

		}
	}
	if (!window.TextEncoder) {
		window.TextEncoder = function (flag, data) {
			if (globalThis.strToUtf8) {
				this.encode = function (data) {
					if ((!data) || data.length === 0) {
						return new Uint8Array([]);
					}
					let buffer = globalThis.strToUtf8(data);
					return new Uint8Array(buffer);
				}
			} else {
				this.encode = function (data) {
					return utf8Encode(data);
				}
			}

		}
	}
})()


// 只能用var
var _$pi = self._$pi = _$pi || (function () {

	// ============================================ 数据结构

	// 所有高层请求的 等待数组（因为这一次 时间不够，或者 其依赖的模块还没有定义
	let _requireArray = [];

	/**
	 * 高层调用require的等待数据
	 */
	class RequireData {
		// reqFunc = ([names对应模块的exports]) => void
		constructor(names, reqFunc = emptyFunc) {
			this.names = names; // 待求模块名数组
			this.reqFunc = reqFunc;
			this.dependSet = new Set(); // 所有模块的依赖数组
			this.nonDefinedSet = new Set(); // key = string，未定义的模块名集合
			this.nonDefinedModuleCount = 0; // 等到为0的时候，重新触发一次getAllDepend
		}

		// 尝试执行
		// 如果依赖全部就绪，就执行；
		// 执行所有模块之后，调用 reqFunc
		// 返回：是否已经执行reqFunc成功；
		tryExec(isAsync = true, accessedSet = null) {
			dependAll(this.names, this.dependSet, this.nonDefinedSet);
			this.nonDefinedModuleCount = this.nonDefinedSet.size;

			if (this.nonDefinedModuleCount > 0) {
				// 等到 defineMod 调用，就会 递减；
				return false;
			}

			// 执行this.names的所有模块
			this.execModules(isAsync, accessedSet);
			return true;
		}

		// 当name模块定义之后的调用；
		// 返回：是否已经执行reqFunc成功；
		defineMod(name) {
			if (this.nonDefinedModuleCount === 0) {
				return true;
			}

			// 如果这里不等 name，那就返回
			if (!this.nonDefinedSet.has(name)) {
				return false;
			}

			this.nonDefinedModuleCount -= 1;
			if (this.nonDefinedModuleCount > 0) {
				// 还有模块没有定义，就返回；
				return false;
			}

			let arr = [...this.nonDefinedSet];
			// 因为 this.nonDefinedModuleCount === 0，就意味着上一批待依赖模块已经就绪；
			this.nonDefinedSet.clear();

			// 从上次没有定义的依赖模块开始搜索新的未定义的依赖模块；
			dependAll(arr, this.dependSet, this.nonDefinedSet);
			this.nonDefinedModuleCount = this.nonDefinedSet.size;

			if (this.nonDefinedModuleCount > 0) {
				// 有新的未定义等待模块，就返回还没就绪；
				return false;
			}

			// 全部就绪，开始执行；
			// 等待模块定义的，都是异步执行的；
			this.execModules(true);
			return true;
		}

		// 执行模块
		// isAsync：意思是 是否可以异步调度；
		execModules(isAsync = true, accessedSet = null) {
			if (isAsync) {
				execSchedule(this.names, this.reqFunc, accessedSet);
			} else {
				execImm(this.names, this.reqFunc, accessedSet);
			}
		}
	}

	// 调度 执行
	// 如果再登录界面中，一下涌入很多app-b app-c的模块，会影响登录界面的体验；
	// 所以要用 exec来缓解下调度策略；
	const execSchedule = (names, reqFunc, accessedSet = null) => {
		pi.exec(() => {
			execImm(names, reqFunc, accessedSet);
		});
	}

	// 立即执行
	const execImm = (names, reqFunc, accessedSet = null) => {
		let mods = [];
		for (let name of names) {
			execModule(name, accessedSet);
			let m = getModule(name);
			mods.push(m.exports);
		}

		let func = reqFunc;
		func(...mods);
	}

	// 用 广度优先遍历 找到所有的依赖
	// moduleNames: string[];  // 开始搜索的模块名
	// dependSet: Set<string>; // 所有modulesNames的依赖的模块组成的集合（包含names本身）；
	// nonDefinedSet: Set<string> // dependSet中，尚未定义的集合；
	const dependAll = (moduleNames, dependSet, nonDefinedSet) => {
		let arr = [...moduleNames];

		let start = 0;
		let end = arr.length;
		// 依次加载数组中的依赖，直到数组长度不再增长
		while (start < end) {
			for (let i = end - 1; i >= start; --i) {
				let name = arr[i];
				dependSet.add(name);

				let mod = getModule(name);

				if (!mod.isDefined()) {
					// 如果还没有定义过，扔到未定义集合中
					nonDefinedSet.add(name);
				} else if (!mod.isBuild()) {
					for (let cname of mod.childrenNames) {
						if (!dependSet.has(cname)) {
							// 将未添加的子模块，扔到arr，进行下一轮遍历
							arr.push(cname);
						}
					}
				}
			}
			// 更新下一轮的 开始 和 结束
			start = end;
			end = arr.length;
		}
	}

	class Module {
		/**
		 * @param {Function} buildFunc = (require, exports, modules, a, b, c, ...) => { ... }
		 * @param {string[]} dependNames 依赖模块名数组
		 */
		constructor(name) {

			this.name = name;
			this.isWasmModule = undefined; // 是否wasm封装模块
			this.wasmModule = undefined; // WebAssemblyModule

			this.isDef = undefined; // 未定义状态

			// 当模块还没有定义，exports为undefined
			this.exports = undefined; // 导出对象

			// 当构建成功后，下面变量全设置为undefined
			this.children = undefined; // 子模块
			this.buildFunc = undefined; // 构建函数
			this.childrenNames = undefined; // 子模块名
		}

		/**
		 * 定义
		 */
		setDefine(func, childrenNames) {
			this.isDef = true;

			this.exports = this.exports || {};
			this.buildFunc = func;

			this.childrenNames = childrenNames;
			this.children = new Array(childrenNames.length);
		}

		/**
		 * 构建成功
		 */
		setBuild() {
			this.buildFunc = undefined;
			this.children = undefined;
			this.childrenNames = undefined;
		}

		/** 
		 * 该模块是否已经定义
		 */
		isDefined() {
			if (!this.isDef) {
				return false;
			}
			if (!this.isWasmModule) {
				return true;
			}

			// 到这里的模块，都是wasm模块；

			if (this.buildFunc) {
				let buildFunc = this.buildFunc;
				this.buildFunc = undefined;
				this.postRun = () => {
					this.isWasmOK = true;
					console.log("WASM load ok, path = " + this.name);
					this.wasmModule = undefined;
					this.wasmBinary = undefined;
					notifyDefineWhenOK(this);
				}

				getWasmRes(this.name).then(([wasmBinary, wasmModule]) => {
					this.wasmModule = wasmModule;
					this.wasmBinary = wasmBinary;
					this.buildFunc = buildFunc;
					execModule(this.name);
				});
			}
			return this.isWasmOK;
		}

		/** 
		 * 该模块是否已经构建
		 */
		isBuild() {
			if (this.isWasmModule) {
				return this.isWasmOK;
			}

			return this.isDef && this.buildFunc === undefined;
		}
	}

	/**
	 * 模块Map
	 *    键：模块绝对路径
	 *    值：Module对象
	 */
	let _modules = new Map();

	/**
	 * 需要立即加载的模块的后缀名数组，
	 * 目前：只有 struct 需要框架层做
	 */
	let _immLoadSuffixs = ["struct"];

	let emptyFunc = () => { };

	// ============================================ 公有接口


	/**
	 * 模块定义
	 * name：模块的路径，相对与项目目录
	 * depends: 该模块依赖的模块名数组，头三个永远是："require", "exports", "module"
	 * func = (require, exports, module, 模块1-Export, 模块-2Export, ...) => { ... }
	 *    参数长度和depends长度相同。
	 */
	let notifyTimer = null;
	const define = (name, dependNames, func) => {
		const needImmLoad = _immLoadSuffixs.some(v => name.endsWith("." + v));
        if (needImmLoad) {
			require(name, emptyFunc);
		}

		let mod = getModule(name);
		if (mod.isDef) {
			throw new Error(`Module ${name} is already defined`);
		}
		let ds = [];
		// 依赖名称数组的头三个永远都是"require", "expots", "module" 去掉
		for (let i = 3; i < dependNames.length; ++i) {
			// 转成相对于name的绝对路径
			ds.push(relativePath(dependNames[i], name));
		}
		mod.setDefine(func, ds);

		// 子模块添加父亲
		for (let cname of ds) {
			let p = getModule(cname);
			if (p.parentNameSet) {
				p.parentNameSet.add(name);
			}
		}
		mod.isWasmModule = name.endsWith(".wasm");

		// 通知对应的函数处理
		if (!mod.isWasmModule) {
			notifyDefineWhenOK(mod);
		}
	}

	// 统一进行定义的处理
	// debug版本是 每个文件 要到服务器请求
	// release版本是 批量合并 所有的 define
	// 如果每个define都马上处理，会多很多未定义模块
	// 所以，debug版本防抖60ms，release版本将处理延迟到promise去，会让一批define先定义完；
	const notifyDefineWhenOK = (mod) => {
		_notifySrc.add(mod.name);

		if (getDebugMode()) {
			// debug，文件不合并，防抖60ms
			if (notifyTimer) {
				clearTimeout(notifyTimer);
				notifyTimer = null;
			}

			notifyTimer = setTimeout(() => {
				notifyTimer = null;
				notifyDefine();
			}, 60);
		} else {
			// release版本，在每个文件的第一个define时候，做一次promise；
			// 等改合并后的文件都define完毕，再统一处理；
			// 这样会多很多已经定义的模块
			if (_notifySrc.size === 1) {
				Promise.resolve().then(() => notifyDefine());
			}
		}
	}

	/**
	 * 模块加载
	 * names: Array or string, 模块名
	 * func = (模块1-Export, 模块-2Export, ...) => { ... }
	 *    参数长度和names长度相同
	 */
	const require = (names, func, currModName, accessedSet = null) => {
		if (!Array.isArray(names)) {
			names = [names];
		}

		// debug模式等待15秒，release模式等待10秒；
		let waitTime = getDebugMode() ? 15 * 1000 : 10 * 1000;
		printWaitModules(waitTime);

		for (let i = 0; i < names.length; ++i) {
			names[i] = relativePath(names[i], currModName);
		}

		let requireData = new RequireData(names, func);
		if (!requireData.tryExec(true, accessedSet)) {
			_requireArray.push(requireData);
		}
	}

	/**
	 * 同步模块加载
	 * names: string | string[], 模块名数组
	 * 返回：如果 参数是字符串，返回模块导出；如果参数是数组，返回模块导出数组
	 */
	const _import = (names, currModName, accessedSet = null) => {

		// debug模式等待15秒，release模式等待10秒；
		let waitTime = getDebugMode() ? 15 * 1000 : 10 * 1000;
		printWaitModules(waitTime);

		let isParamString = typeof (names) === typeof ("");
		if (isParamString) {
			names = [names];
		}

		for (let i = 0; i < names.length; ++i) {
			names[i] = relativePath(names[i], currModName);
		}

		let execMods = undefined;
		let requireData = new RequireData(names, (...exports) => {
			execMods = exports;
		});

		requireData.tryExec(false, accessedSet);
		if (requireData.nonDefinedSet.size > 0) {
			throw new Error(`exec mod failed, waits = ${[...requireData.nonDefinedSet]}`);
		}
		return isParamString ? execMods[0] : execMods;
	}

	// ============================================ 本地方法


	let webAssemblyisinit = false;
	// 获取wasm资源，返回 Promise<[Uint8Array, WasmModule]>
	const getWasmRes = (wasmPath) => {
		if (window.WXWebAssembly && window.WXWebAssembly.instantiate) {
			if (!webAssemblyisinit) {
				// 适配ios的wasm加载
				window.WebAssembly = window.WebAssembly || {};
				window.WebAssembly.Memory = window.WXWebAssembly.Memory;
				window.WebAssembly.Table = window.WXWebAssembly.Table;

				// 将window.WebAssembly.Table赋值给Instance仅仅时为了让Instance不为undefined
				// wasm-bindgen编译出来的wasm（如resmrg）在初始化实例后，会使用instanceof来判断是否为WebAssembly.Instance类型
				// 正好，window.WXWebAssembly.instantiate的返回值不为WebAssembly.Instance类型，所以希望instanceof判断失败

				// TODO: 2021.08.18 号更新，iOS 微信 最新版本 从 8.0.9 更新到 8.0.10 之后，wasm_engine.wasm.ts 中：load函数的 instanceof WebAssembly.Table 直接崩溃，报错：instanceof object 不支持，为了能让 项目能进去，根据上面注释的描述，改为：赋值成 function() {}，结果能进去
				// window.WebAssembly.Instance = window.WebAssembly.Table;
				window.WebAssembly.Instance = function () { };

				window.WebAssembly.instantiate = (buffer, importobj) => {
					let path = new TextDecoder("utf8").decode(buffer);
					return window.WXWebAssembly.instantiate(path, importobj);
				}
				webAssemblyisinit = true;
			}
			return Promise.resolve().then(() => {
				let buffer = new TextEncoder("utf8").encode(wasmPath).buffer;
				return [buffer, buffer];
			})
		} else if (window.qg && window.loadRuntime) {
			if (!webAssemblyisinit) {
				// WebAssembly相关测试于2022.12.28 yuq
				// 华为快游戏WebAssembly的实现有限,而且有些实现还有问题,经测试,华为快游戏必须使用以下方式加载wasm
				// wasmPath必须是本地路径而且只能是本地路径
				// loadRuntime().instantiate(wasmPath, importObject).then(r => {
				// 	r.instance.exports.exported_func();
				// })
				// 并且wasmPath必须是本地的路径,不能是ArrayBuffer
				window.WebAssembly.Instance = function () { };    // 没有实现Instance
				// WebAssembly.instantiate 本身有实现,但是测试发现无法正确加载wasm,这里重新赋值
				window.WebAssembly.instantiate = loadRuntime().instantiate;
				// WebAssembly.compile 无法正确编译
			}
			return Promise.resolve([wasmPath, wasmPath]);
		}

		let {
			Depend
		} = _import("pi_sys/util/depend", "");
		let {
			loadRes
		} = _import("pi_sys/load/res_load", "");

		if (!Depend || !loadRes) {
			throw new Error("must get module after sys init");
		}

		let binaryData = undefined;

		// 加载wasm资源的同时，将wasm的module放到模块约定的一个变量上；
		return loadRes(Depend.getFile(wasmPath)).then((wasmData) => {
			// 有可能是 Uint8Array 或者 Module
			if (WebAssembly.Module && wasmData instanceof WebAssembly.Module) {
				return wasmData;
			} else {
				binaryData = wasmData;
				return WebAssembly.compile(wasmData);
			}
		}).then(wasmModule => {
			// 2020.07.03 经测试，目前chrome不支持IndexedDB存储WebAssembly.Module，所以暂时注释这段代码

			// try {
			//     FileLoadParam.localStore.write(wasmPath, wasmModule);
			// } catch (e) {

			// }

			return [binaryData, wasmModule];
		});
	}

	const getModule = (name) => {
		if (!_modules.has(name)) {
			_modules.set(name, new Module(name));
		}
		return _modules.get(name);
	}

	let _notifySrc = new Set();
	const notifyDefine = () => {
		for (let name of _notifySrc) {
			for (let i = 0; i < _requireArray.length;) {
				let requireData = _requireArray[i];
				if (!requireData.defineMod(name)) {
					++i;
				} else {
					// 特殊技巧将requireData从数组移除
					if (i + 1 < _requireArray.length) {
						_requireArray[i] = _requireArray[_requireArray.length - 1];
					}
					_requireArray.length -= 1;
				}
			}
		}
		_notifySrc.clear();
	}

	/** 
	 * 执行模块, waits = Set<string>();
	 * 返回等待模块的数量
	 */
	const execModule = (name, accessedSet = null) => {
		let mod = getModule(name);

		if (mod.exports === undefined) {
			throw new Error(`Module ${name} isn't defined`);
		}

		// 已经加载
		if (mod.isBuild()) {
			return;
		}

		if (!accessedSet) {
			accessedSet = new Set();
		}

		// 子模块
		accessedSet.add(name);
		for (let i = 0; i < mod.childrenNames.length; ++i) {
			if (!mod.children[i]) {
				let cname = mod.childrenNames[i];

				// 循环，则直接指定
				let c = _modules.get(cname);
				if (accessedSet.has(cname)) {
					mod.children[i] = c.exports;
					continue;
				}

				execModule(cname, accessedSet);
				if (c.isBuild()) {
					mod.children[i] = c.exports;
				} else {
					throw new Error(`build module ${cname} failed`);
				}
			}
		}
		accessedSet.delete(name);

		let func = mod.buildFunc;
		mod.buildFunc = undefined;

		const reqFunc = (names, f) => {
			if (f) {
				require(names, f, name, accessedSet);
			} else {
				return _import(names, name, accessedSet);
			}
		}
		try {
			func(reqFunc, mod.exports, mod, ...mod.children);
		} catch (error) {
			console.log("execModule fail,", error.message, ", mod: ", name);
		}

		if (!mod.isWasmModule) {
			mod.setBuild();
		}
	}

	/**
	 * 以 dir的目录路径 为基准，计算filePath的绝对路径。
	 * 注：如果filePath不以 . 开头，那么filePath就是绝对路径，直接返回即可。
	 */
	const relativePath = (filePath, dir) => {
		// 不以 . 开头，就是绝对路径，直接返回
		if (filePath.charCodeAt(0) !== 46) {
			return filePath;
		}

		let i = 0;
		let len = filePath.length;
		let j = dir.length - 1;

		// 最后一个字符不是/，就代表dir不是目录，需要定位到目录
		if (dir.charCodeAt(j) !== 47) {
			j = dir.lastIndexOf("/");
		}

		while (i < len) {
			if (filePath.charCodeAt(i) !== 46) {
				break;
			}
			if (filePath.charCodeAt(i + 1) === 47) { // ./的情况
				i += 2;
				break;
			}
			if (filePath.charCodeAt(i + 1) !== 46 || filePath.charCodeAt(i + 2) !== 47) {
				break;
			}
			// ../的情况
			i += 3;
			j = dir.lastIndexOf("/", j - 1);
		}

		if (i > 0) {
			filePath = filePath.slice(i);
		}

		if (j < 0) {
			return filePath;
		}

		if (j < dir.length - 1) {
			dir = dir.slice(0, j + 1);
		}

		return dir + filePath;
	}

	const getEnv = () => {
		let m = _$pi._modules.get("pi_sys/util/env");
		if (m && m.exports) {
			return m.exports.Env;
		}
		return undefined;
	}
	/**
	 * 是否debug模式，返回：true，false或者undefined
	 */
	let isDebugMode = undefined;
	const getDebugMode = () => {
		if (isDebugMode === undefined) {
			let env = getEnv();
			isDebugMode = env && env.get("debug");
		}
		return isDebugMode;
	}

	/**
	 * 打印等待的模块
	 * 技巧：debounce
	 */
	let lastWaitTimeout = undefined;
	const printWaitModules = (waitTime) => {
		if (lastWaitTimeout) {
			clearTimeout(lastWaitTimeout);
			lastWaitTimeout = undefined;
		}

		lastWaitTimeout = setTimeout(() => {

			lastWaitTimeout = undefined;

			// 打印一次，将所有的 _requireArray打印出来即可
			for (let rd of _requireArray) {
				let message = {
					names: [...rd.names],
					waits: [...rd.nonDefinedSet],
				};
				console.error(`require module fail, module: ${JSON.stringify(message)}`);
			}
		}, waitTime);
	}

	// 判断wasm是否可用
	let _canUseWasm = undefined;
	const canUseWasm = () => {

		if (_canUseWasm !== undefined) {
			return _canUseWasm;
		}

		let platformType = getPlatformType();

		// TODO 判断平台
		if (window["qg"]) {
			_canUseWasm = false;
			return _canUseWasm;
		}
		console.log("platformType", platformType);
		if (
			platformType !== "web" &&
			platformType !== "webview" &&
			platformType !== "android" &&
			platformType !== "ios" &&
			platformType !== "windows" &&
			platformType !== "wx_minigame" &&
			platformType !== "qq_minigame") {
			_canUseWasm = false;
		} else {
			let wasm = window["WebAssembly"] || window["WXWebAssembly"];
			console.log("wasm", wasm, window["WXWebAssembly"], _canUseWasm);

			// 这里 不判断Stream相关的函数，因为：我们没用到，而且有些Android浏览器不支持
			_canUseWasm = (!!wasm) && (!!wasm.instantiate)
			console.log("wasm", wasm, window["WXWebAssembly"], _canUseWasm);
		}
		return _canUseWasm;
	}

	// 因为模块系统是通用的，所以需要这里判断平台；
	const getPlatformType = () => {
		let type = "web";
		let agent = "";
		try {
			agent = navigator.userAgent || "";
		} catch (e) {
			agent = "";
			self.navigator = {
				userAgent: agent
			}
		}

		agent = agent.toLowerCase();
		if (agent === "") {
			// 一定是小游戏平台
			try {
				let _ = !!qq;
				type = "qq_minigame";
			} catch (_) {
				// TODO: 需要继续判断其他小游戏平台
				type = "wx_minigame";
			}
		} else if (agent.indexOf("pi_webview") >= 0) {
			type = "webview";
		} else if (agent.indexOf("pi_android") >= 0) {
			type = "android";
		} else if (agent.indexOf("pi_ios") >= 0) {
			type = "ios";
		} else if (agent.indexOf("pi_windows") >= 0) {
			type = "windows";
		} else if (agent.indexOf("wechatdevtools") >= 0) {
			type = "wx_minigame";
		} else if (agent.indexOf("qqdevtools") >= 0) {
			type = "qq_minigame";
		}
		return type;
	}


	// ====================== 返回

	const pi = {
		define,
		canUseWasm,

		// 一般会在 app_a 加载 app_b 之前，将 task的set方法换掉这个
		// 目的是为了 让 app-b，c的模块能 调度，不至于卡住 界面；
		// 代码如下：
		//       import { set } from "pi_utils/util/task_mgr";
		//       PISYS.Module.setModuleScheduler(set); // 这里的实现就是: _$pi.exec = set; 
		exec(func = () => { }) { // 用于给使用层修改成task-pool的exec；
			return func()
		},
		require(names, func) {
			if (func) {
				// 异步环境
				pi.exec(() => require(names, func, ""));
			} else {
				// 同步环境，直接执行并返回 exports
				return _import(names, "");
			}
		},
		_modules, // 用于在控制台打印模块的基本信息，如果项目想要加载模块，请用 "pi_sys/util/module"
		_requireArray, // 用于在控制台打印模块的基本信息；
		_immLoadSuffixs, // 如果项目想要设置立即执行的模块后缀，请用 "pi_sys/util/module"
	};
    exports._$pi = pi;
	return pi;
}());