import { HttpRequest, HttpResponse } from "pi_pt/init/httprpc";
import { getPidCount, getPidList, getVmMemory } from "pi_pt/mgr/mgr_lib";
import { get_pid_statistics } from "pi_pt/native/pi_serv_lib/js_lib";
import { SysInfo } from "pi_pt/native/pi_serv_lib/js_sysinfo";
import { readDb } from "pi_pt/util/metric";
import { sendMsg } from "pi_pt/util/pid_rpc";
import { getExtGray, getGrayV, getVids, get_pid, to_pid } from "pi_pt/util/util";
import { checkHttpBasicAuth } from "./auth_lib";
import { initProjectDef } from "pi_pt/init/load_project";
import { checkHttpPortAuth, getEnv } from "./http_proxy.event";
import { checkAuth } from "./user.event";
import { AUTH, FM, RESULT_NUM } from "../../../constant";

declare var readFile;

// #[httpRpc=/mgr/get_mem]
export const mem = (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";
    return getVmMemory(req, resp);
}

// 获取所有虚拟机
// #[httpRpc=/mgr/get_vids]
export const vids = (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";

    return {
        workerVm: Array.from(getVids(getGrayV())),
        extWorkerVm: Array.from(getVids(getExtGray()))
    }
}

// #[httpRpc=/mgr/get_pids]
export const pids = (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";
    return getPidList(req, resp);
}

// #[httpRpc=/mgr/get_pid_count]
export const pid_count = (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";
    return getPidCount(req, resp);
}

// 设置堆栈采集开关
// #[httpRpc=/mgr/stackOnOff]
export const stackOnOff = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";
    let data = req.getData();
    const vid = data.vid;
    const cid = data.cid;
    const off = data.off;
    if (!vid) return "vid null";
    if (!cid) return "cid null";
    if (!off) return "off null";
    const pid = to_pid(Number(vid), Number(cid));
    console.log("pid:", pid);
    return await sendMsg(pid, 'stackOnOff', (off == "true") ? true : false);
}

// 获取指定pid的堆栈信息
// #[httpRpc=/mgr/get_pid_stack]
export const get_pid_stack = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";
    console.log("当前pid:", get_pid());
    let data = req.getData();
    const vid = data.vid;
    const cid = data.cid;
    if (!vid) return "vid null";
    if (!cid) return "cid null";
    const pid = to_pid(Number(vid), Number(cid));
    let r = get_pid_statistics(pid);
    return r;
}

// 开启堆栈打印采集
//#[event=stackOnOff, src=pid_rpc]
export const setStackOnOff = async (_eventName: string, arg: any) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    if (typeof arg == "boolean") {
        (self).stackOnOff = arg;
        return "ok";
    }

    return "error";
}

// 获取负载数据
// #[httpRpc=/mgr/get_load]
export const get_load = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.R, localEnv)) throw "权限错误";
    return await readDb();
}

// gc指定虚拟机
// #[httpRpc=/mgr/gc]
export const vmgc = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.W, localEnv)) throw "权限错误";
    // 获取vid
    const vid = req.getData().vid;
    const path = req.getData().path;
    if (!vid) return "vid null";
    const pid = BigInt(Number(vid)) << BigInt(32);
    ESProcess.send(Number(pid), ["gc"]);
    return "ok"
}


// 指定虚拟机快照
// #[httpRpc=/mgr/heapSnapshot]
export const heapSnapshot = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.W, localEnv)) throw "权限错误";
    // 获取vid
    const vid = req.getData().vid;
    const path = req.getData().path;
    if (!vid) return "vid null";
    (<any>performance).heapSnapshot(path || "./", Number(vid));
    return "抓取快照完成,请在服务器对应目录中查看"
}

// 启动新版本
// #[httpRpc=/mgr/loadProject]
export const loadProject = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.PT_LOAD_INFO, AUTH.W, localEnv)) throw "权限错误";
    const version = req.getData().version;
    if (!version) return "version null";
    // 启动的版本必须大于lastversion
    let lastversion = readFile("lastversion", false);
    if (Number(version) > Number(lastversion)) {
        await initProjectDef(version);
        return "ok"
    } else {
        return `error version, lastversion:${lastversion}`;
    }

}